/*
 * Linux PAM example utility.
 *
 * Copyright, 2010 Grisha Sitkarev
 * <sitkarev@unixkomi.ru>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <termios.h>
#include <security/pam_appl.h>

#define ARRAY_SIZE(a)	(sizeof(a)/sizeof(a[0]))

static void
echo_on(int ttyfd)
{
	struct termios tio;
	int res;

	res = tcgetattr(ttyfd, &tio);
	if (res == -1)
		perror("tcgetattr()");
	tio.c_lflag |= ECHO;
	res = tcsetattr(ttyfd, TCSANOW, &tio);
	if (res == -1)
		perror("tcsetattr()");
}

static void
echo_off(int ttyfd)
{
	struct termios tio;
	int res;

	res = tcgetattr(ttyfd, &tio);
	if (res == -1)
		perror("tcgetattr()");
	tio.c_lflag &= ~ECHO;
	res = tcsetattr(ttyfd, TCSANOW, &tio);
	if (res == -1)
		perror("tcsetattr()");
}

static char *
getline_chomp_alloc(FILE *fp)
{
	char *str;
	int nalloc, nstr;
	int gotlf;

	str = NULL;
	nstr = nalloc = gotlf = 0;

	while (!gotlf) {
		int n;
		char *tmp;
		char *s, buf[1024];

		s = fgets(buf, sizeof(buf), fp);
		if (s == NULL)
			goto error;
		n = strlen(s);
		tmp = realloc(str, nalloc+sizeof(buf));
		if (tmp == NULL)
			goto error;
		str = tmp;
		nalloc += sizeof(buf);
		if (n > 0 && buf[n-1] == '\n') {
			buf[--n] = '\0';
			gotlf = 1;
		} else if (n < sizeof(buf)-1) {
			gotlf = 1;
		}
		memcpy(str+nstr, buf, n);
		nstr += n;
		str[nstr] = '\0';
	}

	if (str != NULL) {
		char *tmp;
		tmp = realloc(str, nstr+1);
		if (tmp == NULL)
			goto error;
		str = tmp;
	}

	return str;
error:
	if (str != NULL)
		free(str);
	return NULL;
}

int
conversation(int num_msg, const struct pam_message **msg,
	     struct pam_response **resp, void *app_data)
{
	struct pam_response *rp;
	const struct pam_message *mp;
	char *str;
	int i, n;

	rp = NULL;
	mp = *msg;

	for (i = 0; i < num_msg; i++) {
		switch (mp[i].msg_style) {
		case PAM_PROMPT_ECHO_ON:
			printf("%s", mp[i].msg);
			fflush(stdout);
			/* release previous response */
			if (rp != NULL) {
				if (rp->resp != NULL)
					free(rp->resp);
				free(rp);
			}
			rp = malloc(sizeof(*rp));
			if (rp == NULL)
				goto error;
			memset(rp, 0, sizeof(*rp));
			str = getline_chomp_alloc(stdin);
			if (str == NULL)
				goto error;
			rp->resp = str;
			break;
		case PAM_PROMPT_ECHO_OFF:
			printf("%s", mp[i].msg);
			fflush(stdout);
			if (rp != NULL) {
				if (rp->resp != NULL)
					free(rp->resp);
				free(rp);
			}
			rp = malloc(sizeof(*rp));
			if (rp == NULL)
				goto error;
			memset(rp, 0, sizeof(*rp));
			echo_off(STDIN_FILENO);
			str = getline_chomp_alloc(stdin);
			echo_on(STDIN_FILENO);
			putc('\n', stdout);
			if (str == NULL)
				goto error;
			rp->resp = str;
			break;
		case PAM_ERROR_MSG:
			n = strlen(mp[i].msg);
			if (n > 0)
				printf("%s%s", mp[i].msg,
				       mp[i].msg[n-1] == '\n' ? "" : "\n");
			break;
		case PAM_TEXT_INFO:
			n = strlen(mp[i].msg);
			if (n > 0)
				printf("%s%s", mp[i].msg,
				       mp[i].msg[n-1] == '\n' ? "" : "\n");
			break;
		}
	}

	*resp = rp;

	return 0;
error:
	if (rp != NULL) {
		if (rp->resp != NULL)
			free(rp->resp);
		free(rp);
	}
	return -1;
}

int
main(int argc, char *argv[])
{
	int opt, res, i;
	char *service;
	char *user;
	struct pam_conv pconv;
	int itemlist[] = { PAM_USER, PAM_TTY, PAM_SERVICE, PAM_AUTHTOK };
	char *itemname[] = { "PAM_USER", "PAM_TTY", "PAM_SERVICE", "PAM_AUTHTOK" };
	char *item;
	pam_handle_t *pamh;

	if (!isatty(STDIN_FILENO) || !isatty(STDOUT_FILENO)) {
		fprintf(stderr, "not on a tty\n");
		exit(1);
	}

	service = "login";
	user = NULL;
	opterr = 0;

	while ((opt = getopt(argc, argv, "s:u:h")) > 0) {
		switch (opt) {
		case 's':
			service = optarg;
			break;
		case 'u':
			user = optarg;
			break;		
		default:
			printf("Usage: %s [-s service] [-u user]\n", argv[0]);
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;

	pconv.conv = conversation;
	pconv.appdata_ptr = NULL;

	res = pam_start(service, user, &pconv, &pamh);

	if (res != PAM_SUCCESS) {
		fprintf(stderr, "pam_start: %s\n", pam_strerror(pamh, res));
		exit(1);
	}

	pam_set_item(pamh, PAM_TTY, ttyname(STDIN_FILENO));

	res = pam_authenticate(pamh, PAM_DISALLOW_NULL_AUTHTOK);

	if (res != PAM_SUCCESS) {
		fprintf(stderr, "pam_authenticate: %s\n", pam_strerror(pamh, res));
		goto end;
	}

	printf("Authentication: %s\n", pam_strerror(pamh, res));

	res = pam_acct_mgmt(pamh, PAM_DISALLOW_NULL_AUTHTOK);

	if (res != PAM_SUCCESS) {
		fprintf(stderr, "pam_acct_mgmt: %s\n", pam_strerror(pamh, res));
		goto end;
	}

	printf("Account: %s\n", pam_strerror(pamh, res));

	for (i = 0; i < ARRAY_SIZE(itemlist); i++) {
		res = pam_get_item(pamh, itemlist[i], (const void **)&item);
		if (res == PAM_SUCCESS) {
			printf("%s: %s\n", itemname[i],
			       item != NULL ? item : "NULL");
		}
	}

end:
	res = pam_end(pamh, 0);

	if (res != PAM_SUCCESS) {
		fprintf(stderr, "pam_end: %s\n", pam_strerror(pamh, res));
		exit(1);
	}

	return 0;
}

